/* Purpleview - simple image viewer
 *
 * Copyright (c) 2002-2003 Jason Summers <jason1@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  See the file purpleview.txt for more information.
 *
 */

#include <windows.h>
#include <tchar.h>
#include <commdlg.h>
#include <commctrl.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>

#include "resource.h"

#include "pvconf.h"
#include "purpleview.h"
#include <strsafe.h>

static char g_software_name_full_a[] = "Purpleview v" PVVERSION_A;

static TCHAR g_software_name[] = _T("Purpleview");
static TCHAR g_software_name_full[] = _T("Purpleview v") PVVERSION;


struct globals_struct {
	HINSTANCE hInst;
	LPBITMAPINFOHEADER lpdib;
	int lpdib_size, lpdib_bits_offs;
	void *lpbits;

	int bordercolorflag; // 0=system default, 1=custom bgcolor
	int fixratio;
	int densityunits;   // 0=meters, 1=cm, 2=inches
	int savescrollpos;
	int stretchmode;
	int shrinkmode;
	int halign,valign;   // 0=left/top,1=center,2=right/bottom
	int xscrollpos, yscrollpos;
	int xdesiredscrpos, ydesiredscrpos;
	int stretchbltmode;

	int dragging;
	int dragpos_x, dragpos_y;
	POINTS mousepos;
	COLORREF orig_custom_bg;
	HCURSOR cur_drag1, cur_drag2;

	double tmp_displayed_gamma;

	COLORREF custcolors[16];
	TCHAR image_filename[MAX_PATH];
};

static struct globals_struct gl;

// information about the original image
struct iinfo_struct * iinfo;

// information about the program environment
struct einfo_struct * einfo;

static const TCHAR *colortype_name[] = {
	_T("Grayscale"),_T(""),_T("RGB"),_T("Palette"),_T("Grayscale+Alpha"),_T(""),_T("RGBA"),_T(""),
	_T("Unknown"),_T("YCbCr"),_T("CMYK"),_T("YCCK")
};

static const TCHAR *compr_name[] = {
	_T("None"),_T("RLE8"),_T("RLE4"),_T(""),_T(""),_T("Deflate"),_T(""),_T("JPEG"),_T("LZW"),_T("Unknown")
};

static const TCHAR *fileformat_name[] = { _T(""),_T("PNG"),_T("BMP"),_T(""),_T("JFIF (JPEG)"),_T("TIFF"),_T("GIF") };


static LRESULT CALLBACK WndProcMain(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK WndProcGammaDisplay(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcAbout(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcInfo(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcAdvOptions(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcSetGamma(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);


void pv_mesg(HWND hwnd,int severity, const TCHAR *fmt, ...)
{
	va_list ap;
	TCHAR *t;
	int flags;
	TCHAR title[100];
	TCHAR buf[3072];

	va_start(ap, fmt);
	StringCbVPrintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);

	switch(severity) {
	case MSG_S: t=_T("Error");   flags=MB_ICONERROR;        break;
	case MSG_W: t=_T("Warning"); flags=MB_ICONWARNING;      break;
	case MSG_I: t=_T("Notice");  flags=MB_ICONINFORMATION;  break;
	default:    t=_T("Error");   flags=MB_ICONWARNING;      break;  // (MSG_E)
	}

	StringCbPrintf(title,sizeof(title),_T("%s - %s"),t,einfo->software_name);
	SetForegroundWindow(hwnd);
	MessageBox(hwnd,buf,title,MB_OK|flags);
}


static void SetTitle(HWND hwnd)
{
	int i, baseindex;
	TCHAR buf[1000];

	if(gl.lpdib) {
		// find base part of filename
		baseindex=0;
		for(i=0;iinfo->img_filename[i];i++) {
			if(iinfo->img_filename[i]=='\\') baseindex=i+1;
		}
		StringCbPrintf(buf,sizeof(buf),_T("%s - %s"),&iinfo->img_filename[baseindex],einfo->software_name);
	}
	else {
		StringCbPrintf(buf,sizeof(buf),_T("%s"),einfo->software_name);
	}

	if(!einfo->use_gamma) {
		StringCbCat(buf,sizeof(buf),_T(" [gamma off]"));
	}
	SetWindowText(hwnd,buf);
}

static void CopyToClipboard(HWND hwnd,void *mem,int size,UINT format)
{
	HGLOBAL hClip;
	LPVOID lpClip;

	if(!mem) return;

	if(!OpenClipboard(NULL)) {
		pv_mesg(hwnd,MSG_E,_T("Can't open the clipboard"));
		return;
	}

	if(EmptyClipboard()) {
		hClip=GlobalAlloc(GMEM_ZEROINIT|GMEM_MOVEABLE|GMEM_DDESHARE,size);
		lpClip=GlobalLock(hClip);
		if(lpClip) {
			CopyMemory(lpClip,mem,size);
			GlobalUnlock(hClip);
			if(!SetClipboardData(format,hClip)) {
				pv_mesg(hwnd,MSG_E,_T("Can't set clipboard data"));
			}
		}
		else {
			pv_mesg(hwnd,MSG_E,_T("Can't allocate memory for clipboard"));
		}
	}
	else {
		pv_mesg(hwnd,MSG_W,_T("Can't clear the clipboard"));
	}
	CloseClipboard();
}


// returns 0 if canceled
static int choose_color_dialog(HWND hwnd, COLORREF *color)
{
	CHOOSECOLOR cc;

	ZeroMemory(&cc,sizeof(CHOOSECOLOR));
	cc.lStructSize=sizeof(CHOOSECOLOR);
	cc.hwndOwner=hwnd;
	cc.rgbResult= *color;
	cc.lpCustColors=gl.custcolors;
	cc.Flags=CC_RGBINIT|CC_FULLOPEN;

	if(ChooseColor(&cc)) {
		(*color) = cc.rgbResult;
		return 1;
	}
	return 0;
}

int pv_get_dib_bytes_per_row(int bpp, int width)
{
	return ((bpp*width+31)/32)*4;
}

static int get_dib_width(void)
{
	if(!gl.lpdib) return 0;
	if(gl.lpdib->biSize >= 40)
		return (int)gl.lpdib->biWidth;
	return (int) ((BITMAPCOREHEADER*)gl.lpdib)->bcWidth;
}

static int get_dib_height(void)
{
	int tmp;
	if(!gl.lpdib) return 0;
	if(gl.lpdib->biSize >= 40) {
		tmp = * (int*) (&gl.lpdib->biHeight);
		if(tmp<0) tmp = -tmp;
		return tmp;
	}
	return (int) ((BITMAPCOREHEADER*)gl.lpdib)->bcHeight;
}

static int get_dib_bpp(void)
{
	if(!gl.lpdib) return 0;
	if(gl.lpdib->biSize >= 40) {
		return (int)gl.lpdib->biBitCount;
	}
	return (int) ((BITMAPCOREHEADER*)gl.lpdib)->bcBitCount;
}

// call after loading a new image (or changing g_fixratio option)
static void calc_image_size(void)
{
	double ratio;

	if(!gl.lpdib) return;

	iinfo->src_width = get_dib_width();
	iinfo->src_height = get_dib_height();

	iinfo->adj_src_width=  iinfo->src_width;
	iinfo->adj_src_height= iinfo->src_height;
	iinfo->adj_src_width_flt= (double)iinfo->src_width;
	iinfo->adj_src_height_flt=(double)iinfo->src_height;

	if(iinfo->res_flag && gl.fixratio) {
		if(iinfo->res_x>0 && iinfo->res_y>0) {

			ratio = (double)iinfo->res_y/(double)iinfo->res_x;

			if(ratio>1.0 && ratio<20.1) {
				iinfo->adj_src_width_flt *= ratio;
				iinfo->adj_src_width= (int)iinfo->adj_src_width_flt;
			}
			else if(ratio<1.0 && ratio>0.049) {
				iinfo->adj_src_height_flt /= ratio;
				iinfo->adj_src_height= (int)iinfo->adj_src_height_flt;
			}
		}
	}
}


static const TCHAR* MYREGKEY = _T("SOFTWARE\\Generic\\Purpleview");

static void save_settings(HWND hwnd)
{
	DWORD disp;
	HKEY key;
	LONG rv;
	DWORD tmp;

	rv=RegCreateKeyEx(HKEY_CURRENT_USER,MYREGKEY,
		0,NULL,REG_OPTION_NON_VOLATILE,KEY_WRITE,
		NULL,&key,&disp);
	if(rv!=ERROR_SUCCESS) return;

	tmp = (gl.shrinkmode<<1)|gl.stretchmode;
	RegSetValueEx(key,_T("stretchmode"),0,REG_DWORD,(BYTE*)&tmp,sizeof(DWORD));
	RegSetValueEx(key,_T("gamma_correct"),0,REG_DWORD,(BYTE*)&einfo->use_gamma,sizeof(DWORD));
	RegSetValueEx(key,_T("use_image_bg"),0,REG_DWORD,(BYTE*)&einfo->use_image_bg,sizeof(DWORD));
	RegSetValueEx(key,_T("use_custom_bg"),0,REG_DWORD,(BYTE*)&einfo->use_custom_bg,sizeof(DWORD));

	RegSetValueEx(key,_T("custom_bg"),0,REG_DWORD,(BYTE*)&einfo->custom_bg,sizeof(einfo->custom_bg));
	RegSetValueEx(key,_T("custcolors"),0,REG_BINARY,(BYTE*)&gl.custcolors,16*sizeof(COLORREF));
	RegSetValueEx(key,_T("border_policy"),0,REG_DWORD,(BYTE*)&gl.bordercolorflag,sizeof(DWORD));
	RegSetValueEx(key,_T("density_units"),0,REG_DWORD,(BYTE*)&gl.densityunits,sizeof(DWORD));

	tmp = (gl.valign<<2)|gl.halign;
	RegSetValueEx(key,_T("align"),0,REG_DWORD,(BYTE*)&tmp,sizeof(DWORD));
	RegSetValueEx(key,_T("fix_ratio"),0,REG_DWORD,(BYTE*)&gl.fixratio,sizeof(DWORD));
	RegSetValueEx(key,_T("save_scroll_pos"),0,REG_DWORD,(BYTE*)&gl.savescrollpos,sizeof(DWORD));
	RegSetValueEx(key,_T("stretchblt_mode"),0,REG_DWORD,(BYTE*)&gl.stretchbltmode,sizeof(DWORD));
	RegSetValueEx(key,_T("bmp_alpha32"),0,REG_DWORD,(BYTE*)&einfo->bmp_alpha32,sizeof(DWORD));

	tmp = (DWORD)(0.5+einfo->monitor_gamma*10000.0);
	RegSetValueEx(key,_T("monitor_gamma"),0,REG_DWORD,(BYTE*)&tmp,sizeof(DWORD));

	RegCloseKey(key);
}

static void read_settings(void)
{
	DWORD datasize,datatype;
	LONG rv;
	HKEY key;
	int i;
	DWORD tmp;

	// init globals
	gl.lpdib=NULL;
	gl.xscrollpos=0; gl.yscrollpos=0;
	gl.xdesiredscrpos=0; gl.ydesiredscrpos=0;

	gl.stretchmode=0; gl.shrinkmode=0;
	gl.halign=gl.valign=1; // center
	gl.fixratio=1;
	gl.bordercolorflag=0;
	gl.densityunits=0;
	gl.savescrollpos=1;
	gl.stretchbltmode = COLORONCOLOR;
	gl.dragging=0;

	einfo->use_image_bg=1;
	einfo->use_custom_bg=1;
	einfo->use_gamma=1;
	gl.orig_custom_bg = RGB(240,230,240);

	einfo->custom_bg = gl.orig_custom_bg;
	iinfo->bgcolor = gl.orig_custom_bg; // ?? (unused)

	iinfo->has_bgcolor = 0;

	iinfo->img_fmt=0;
	iinfo->res_flag=0;

	iinfo->colortype= 7;
	iinfo->bitsperpixel= -1;
	iinfo->interlaced= -1;
	iinfo->palettecolors= -1;
	iinfo->compression=0;

	for(i=0;i<16;i++) gl.custcolors[i] = RGB(128,0,128);

	gl.cur_drag1=LoadCursor(gl.hInst,_T("CUR_DRAG1"));
	gl.cur_drag2=LoadCursor(gl.hInst,_T("CUR_DRAG2"));

	// read registry

	rv=RegOpenKeyEx(HKEY_CURRENT_USER,MYREGKEY,0,KEY_READ,&key);
	if(rv==ERROR_SUCCESS) {

		datasize=sizeof(DWORD);
		if(RegQueryValueEx(key,_T("stretchmode"),NULL,&datatype,(LPBYTE)&tmp,&datasize)==ERROR_SUCCESS) {
			gl.stretchmode = tmp&1;
			gl.shrinkmode = (tmp&2)>>1;
		}

		datasize=sizeof(DWORD);
		RegQueryValueEx(key,_T("gamma_correct"),NULL,&datatype,(LPBYTE)&einfo->use_gamma,&datasize);
		datasize=sizeof(DWORD);
		RegQueryValueEx(key,_T("use_image_bg"),NULL,&datatype,(LPBYTE)&einfo->use_image_bg,&datasize);
		datasize=sizeof(DWORD);
		RegQueryValueEx(key,_T("use_custom_bg"),NULL,&datatype,(LPBYTE)&einfo->use_custom_bg,&datasize);

		datasize=sizeof(einfo->custom_bg);
		RegQueryValueEx(key,_T("custom_bg"),NULL,&datatype,(LPBYTE)&einfo->custom_bg,&datasize);

		datasize=16*sizeof(COLORREF);
		RegQueryValueEx(key,_T("custcolors"),NULL,&datatype,(LPBYTE)&gl.custcolors,&datasize);

		datasize=sizeof(DWORD);
		RegQueryValueEx(key,_T("border_policy"),NULL,&datatype,(LPBYTE)&gl.bordercolorflag,&datasize);

		datasize=sizeof(DWORD);
		RegQueryValueEx(key,_T("density_units"),NULL,&datatype,(LPBYTE)&gl.densityunits,&datasize);

		datasize=sizeof(DWORD);
		RegQueryValueEx(key,_T("fix_ratio"),NULL,&datatype,(LPBYTE)&gl.fixratio,&datasize);
		datasize=sizeof(DWORD);
		RegQueryValueEx(key,_T("save_scroll_pos"),NULL,&datatype,(LPBYTE)&gl.savescrollpos,&datasize);
		datasize=sizeof(DWORD);
		RegQueryValueEx(key,_T("stretchblt_mode"),NULL,&datatype,(LPBYTE)&gl.stretchbltmode,&datasize);

		datasize=sizeof(DWORD);
		RegQueryValueEx(key,_T("bmp_alpha32"),NULL,&datatype,(LPBYTE)&einfo->bmp_alpha32,&datasize);

		datasize=sizeof(DWORD);
		if(RegQueryValueEx(key,_T("align"),NULL,&datatype,(LPBYTE)&tmp,&datasize)==ERROR_SUCCESS) {
			gl.halign = tmp&3;
			gl.valign = (tmp&12)>>2;
		}

		datasize=sizeof(DWORD);
		if(RegQueryValueEx(key,_T("monitor_gamma"),NULL,&datatype,(LPBYTE)&tmp,&datasize)==ERROR_SUCCESS) {
			einfo->monitor_gamma = ((double)tmp)/10000.0;
		}

		RegCloseKey(key);
	}

	gl.custcolors[15] = gl.orig_custom_bg; // You can't customize this one. (hah!)

}


// sets lpdib_bits_offs and lpbits based on lpdib
static void find_dib_bits(void)
{
	int palentries;
	int bitcount;

	if(!gl.lpdib) return;
	if(gl.lpdib->biSize == 12) {
		bitcount = ((LPBITMAPCOREHEADER)gl.lpdib)->bcBitCount;
		palentries=0;
		if(bitcount<=8) {
			palentries = 1<<bitcount;
		}
		gl.lpdib_bits_offs = gl.lpdib->biSize + 3*palentries;
	}
	else {
		bitcount = gl.lpdib->biBitCount;
		palentries = gl.lpdib->biClrUsed;
		if(palentries==0 && bitcount<=8) {
			palentries = 1<<bitcount;
		}
		gl.lpdib_bits_offs = gl.lpdib->biSize + 4*palentries  +
			( (gl.lpdib->biCompression==BI_BITFIELDS)?12:0 );
	}
	gl.lpbits = &(((unsigned char*)gl.lpdib)[gl.lpdib_bits_offs]);
}


// call when window is resized  -- set scrollbar thumb positions
static void set_scrollbars(HWND hwnd)
{
	RECT r;
	SCROLLINFO si;

	if(!gl.lpdib && !gl.shrinkmode) {
		EnableScrollBar(hwnd,SB_HORZ,ESB_DISABLE_BOTH);
		EnableScrollBar(hwnd,SB_VERT,ESB_DISABLE_BOTH);
	}
	if(!gl.lpdib) return;
	if(gl.shrinkmode) return; // no scrollbars in this mode
	
	GetClientRect(hwnd,&r);

	if(iinfo->adj_src_width > r.right) {
		EnableScrollBar(hwnd,SB_HORZ,ESB_ENABLE_BOTH);
		gl.xscrollpos = gl.xdesiredscrpos;
		if( (iinfo->adj_src_width-gl.xscrollpos) < r.right) {
			gl.xscrollpos = iinfo->adj_src_width - r.right;
		}
		ZeroMemory(&si,sizeof(SCROLLINFO));
		si.cbSize=sizeof(SCROLLINFO);
		si.fMask=SIF_ALL|SIF_DISABLENOSCROLL; 
		si.nMin=0;
		si.nPage=r.right;
		si.nMax=iinfo->adj_src_width-1;
		si.nPos= gl.xscrollpos;
		SetScrollInfo(hwnd,SB_HORZ,&si,TRUE);
	}
	else {
		gl.xscrollpos=0;
		EnableScrollBar(hwnd,SB_HORZ,ESB_DISABLE_BOTH);
	}

	if(iinfo->adj_src_height > r.bottom) {
		EnableScrollBar(hwnd,SB_VERT,ESB_ENABLE_BOTH);
		gl.yscrollpos = gl.ydesiredscrpos;
		if( (iinfo->adj_src_height-gl.yscrollpos) < r.bottom) {
			gl.yscrollpos = iinfo->adj_src_height - r.bottom;
		}
		ZeroMemory(&si,sizeof(SCROLLINFO));
		si.cbSize=sizeof(SCROLLINFO);
		si.fMask=SIF_ALL|SIF_DISABLENOSCROLL; 
		si.nMin=0;
		si.nPage=r.bottom;
		si.nMax=iinfo->adj_src_height-1;
		si.nPos= gl.yscrollpos;
		SetScrollInfo(hwnd,SB_VERT,&si,TRUE);
	}
	else {
		gl.yscrollpos=0;
		EnableScrollBar(hwnd,SB_VERT,ESB_DISABLE_BOTH);
	}

	if(!gl.savescrollpos) {
		gl.xdesiredscrpos=gl.xscrollpos;
		gl.ydesiredscrpos=gl.yscrollpos;
	}
}


static void scrollbarmsg_x(HWND hwnd, int code, int pos)
{
	int oldpos, newpos;
	RECT r;
	SCROLLINFO si;
	int page;

	if(gl.shrinkmode) return;
	if(!gl.lpdib) return;

	GetClientRect(hwnd,&r);

	oldpos = gl.xscrollpos;
	newpos = gl.xscrollpos;

	page=r.right-10; if(page<10) page=10;
	switch(code) {
		case SB_LINELEFT:  newpos-=30; break;
		case SB_LINERIGHT: newpos+=30; break;
		case SB_PAGELEFT:  newpos-=page; break;
		case SB_PAGERIGHT: newpos+=page; break;
		case SB_LEFT:      newpos=0; break;
		case SB_RIGHT:     newpos=r.right; break;
		case SB_THUMBPOSITION:
		case SB_THUMBTRACK:
			newpos=pos;
			break;
	}

	if(newpos>(iinfo->adj_src_width-r.right))
		newpos=iinfo->adj_src_width-r.right;
	if(newpos<0) newpos=0;
	if(newpos==oldpos) return; // no change

	gl.xscrollpos = newpos;
	gl.xdesiredscrpos = newpos;

	ZeroMemory(&si,sizeof(SCROLLINFO));
	si.cbSize = sizeof(SCROLLINFO);
	si.fMask = SIF_POS;
	si.nPos = newpos;
	SetScrollInfo(hwnd,SB_HORZ,&si,TRUE);

	ScrollWindowEx(hwnd, oldpos-newpos, 0, NULL, &r, NULL,NULL,SW_INVALIDATE);
}

static void mousedrag(HWND hwnd, int xpos, int ypos)
{
	int oldpos_x, newpos_x;
	int oldpos_y, newpos_y;
	RECT r;
	SCROLLINFO si;

	if(gl.shrinkmode) return;
	if(!gl.lpdib) return;

	GetClientRect(hwnd,&r);

	oldpos_y = gl.yscrollpos;
	newpos_y = ypos;

	oldpos_x = gl.xscrollpos;
	newpos_x = xpos;

	if(newpos_y>(iinfo->adj_src_height-r.bottom))
		newpos_y=iinfo->adj_src_height-r.bottom;
	if(newpos_y<0) newpos_y=0;

	if(newpos_x>(iinfo->adj_src_width-r.right))
		newpos_x=iinfo->adj_src_width-r.right;
	if(newpos_x<0) newpos_x=0;

	if(newpos_y==oldpos_y && newpos_x==oldpos_x) return; // no change

	gl.yscrollpos = newpos_y;
	gl.ydesiredscrpos = newpos_y;
	gl.xscrollpos = newpos_x;
	gl.xdesiredscrpos = newpos_x;

	ZeroMemory(&si,sizeof(SCROLLINFO));
	si.cbSize = sizeof(SCROLLINFO);
	si.fMask = SIF_POS;
	if(newpos_y!=oldpos_y) {
		si.nPos = newpos_y;
		SetScrollInfo(hwnd,SB_VERT,&si,TRUE);
	}
	if(newpos_x!=oldpos_x) {
		si.nPos = newpos_x;
		SetScrollInfo(hwnd,SB_HORZ,&si,TRUE);
	}

	ScrollWindowEx(hwnd, oldpos_x-newpos_x, oldpos_y-newpos_y,  NULL, &r, NULL,NULL,SW_INVALIDATE);
}

static void scrollbarmsg_y(HWND hwnd, int code, int pos)
{
	int oldpos, newpos;
	RECT r;
	SCROLLINFO si;
	int page;

	if(gl.shrinkmode) return;
	if(!gl.lpdib) return;

	GetClientRect(hwnd,&r);

	oldpos = gl.yscrollpos;
	newpos = gl.yscrollpos;

	page=r.bottom-10; if(page<10) page=10;
	switch(code) {
		case SB_LINEUP:  newpos-=30; break;
		case SB_LINEDOWN: newpos+=30; break;
		case SB_PAGEUP:  newpos-=page; break;
		case SB_PAGEDOWN: newpos+=page; break;
		case SB_TOP:       newpos=0; break;
		case SB_BOTTOM:    newpos=r.right; break;
		case SB_THUMBPOSITION:
		case SB_THUMBTRACK:
			newpos=pos;
			break;
	}

	if(newpos>(iinfo->adj_src_height-r.bottom))
		newpos=iinfo->adj_src_height-r.bottom;
	if(newpos<0) newpos=0;
	if(newpos==oldpos) return; // no change

	gl.yscrollpos = newpos;
	gl.ydesiredscrpos = newpos;

	ZeroMemory(&si,sizeof(SCROLLINFO));
	si.cbSize = sizeof(SCROLLINFO);
	si.fMask = SIF_POS;
	si.nPos = newpos;
	SetScrollInfo(hwnd,SB_VERT,&si,TRUE);

	ScrollWindowEx(hwnd, 0, oldpos-newpos,  NULL, &r, NULL,NULL,SW_INVALIDATE);
}


// img_source is normally 0, meaning from a file.
//   Can be 1 for type BMP, to get the file from the clipboard.
// Probably shouldn't need the filesize parameter, but it's there until I
//   can work out a cleaner way to do this.
static int load_image_main(HWND hwnd,int fmt, int img_source, TCHAR *fn,
					unsigned int filesize, int reloading)
{
	int ret;
	struct iinfo_struct new_iinfo;

	ZeroMemory(&new_iinfo,sizeof(struct iinfo_struct));
	new_iinfo.img_source = img_source;
	new_iinfo.img_filename = fn;
	new_iinfo.file_size = filesize;

	ret=0;
	switch(fmt) {
	case FMT_BMP:
		ret=pv_bmp_open(hwnd,&new_iinfo, einfo);
		break;
#if PV_PNG_SUPPORT
	case FMT_PNG:
		ret=pv_png_open(hwnd,&new_iinfo, einfo);
		break;
#endif
#if PV_JPEG_SUPPORT
	case FMT_JPEG:
		ret=pv_jpeg_open(hwnd,&new_iinfo,einfo);
		break;
#endif
#if PV_GIF_SUPPORT
	case FMT_GIF:
		ret=pv_gif_open(hwnd,&new_iinfo,einfo);
		break;
#endif
#if PV_TIFFREAD_SUPPORT
	case FMT_TIFF:
		ret=pv_tiff_open(hwnd,&new_iinfo,einfo);
		break;
#endif
	default:
		ret=0;
	}

	if(ret) {
		//int baseindex,i;

		if(!new_iinfo.lpdib || !new_iinfo.lpdib_size) {
			pv_mesg(hwnd,MSG_S,_T("Image loader failed to set image data"));
			return 0;
		}
		if(gl.lpdib) { HeapFree(einfo->heap,0,gl.lpdib); gl.lpdib=NULL; } // free old image
		gl.lpdib=new_iinfo.lpdib;  //new_lpdib;
		gl.lpdib_size=new_iinfo.lpdib_size; //=new_lpdib_size;
		*iinfo = new_iinfo;    // structure copy

		iinfo->img_filename= gl.image_filename;
		StringCchCopy(iinfo->img_filename,MAX_PATH,fn);

		if(iinfo->lpdib_bits_offs>0) {
			gl.lpdib_bits_offs= iinfo->lpdib_bits_offs;  // fixme. this is ugly
			gl.lpbits = &(((unsigned char*)gl.lpdib)[gl.lpdib_bits_offs]);
		}
		else {
			find_dib_bits();
		}

		calc_image_size();
		if(!reloading) {
			gl.xscrollpos=0; gl.yscrollpos=0;
			gl.xdesiredscrpos=0; gl.ydesiredscrpos=0;
		}

		set_scrollbars(hwnd);

		SetTitle(hwnd);

		InvalidateRect(hwnd,NULL,FALSE);
	}
	return 1;
}


static int save_image_main(HWND hwnd,int fmt)
{
	if(!gl.lpdib) {
		pv_mesg(hwnd,MSG_S,_T("No image to save"));
		return 0;
	}

	iinfo->lpdib = gl.lpdib;
	iinfo->lpdib_size = gl.lpdib_size;
	iinfo->lpdib_bits_offs = gl.lpdib_bits_offs;

	switch(fmt) {
	case FMT_BMP:
		pv_bmp_save(hwnd,iinfo,einfo);
		break;
	case FMT_TIFF:
		pv_tiff_save(hwnd,iinfo,einfo);
		break;
#if PV_PNG_SUPPORT
	case FMT_PNG:
		pv_png_save(hwnd,iinfo,einfo);
		break;
#endif
	default:
		return 0;

	}
	return 1;
}

// Detects the image type and calls the right function to load it.
// This is a little inefficient because it opens and
// closes the image once just to determine its format.
static int open_image_by_name(HWND hwnd, TCHAR *fn)
{
	HANDLE hfile;
	DWORD bytesread;
	BOOL b;
	int ret;
	unsigned char fbuf[20];
	DWORD tmpd;
	unsigned int filesize;

	hfile=CreateFile(fn,GENERIC_READ,FILE_SHARE_READ,NULL,
		OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(hfile==INVALID_HANDLE_VALUE) {
		pv_mesg(hwnd,MSG_E,_T("Can't open %s"),fn);
		return 0;
	}

	// We take this opportunity to get the file size.
	// This is only for display purposes later.
	tmpd=GetFileSize(hfile,NULL);
	if(tmpd==0xffffffff) filesize=0;
	else filesize= (int)tmpd;

	b=ReadFile(hfile,(void*)fbuf,8,&bytesread,NULL);
	CloseHandle(hfile);

	if(!b) {
		pv_mesg(hwnd,MSG_E,_T("Can't read file"));
		return 0;
	}

	if(bytesread<8) {
		fbuf[0]=fbuf[1]= '\0';
	}

	if(fbuf[0]=='B' && fbuf[1]=='M') { // looks like a BMP file.
		load_image_main(hwnd,FMT_BMP,0,fn, filesize, 0);
	}
#if PV_PNG_SUPPORT
	else if(fbuf[0]==137 && fbuf[1]=='P') { // looks like a PNG file
		load_image_main(hwnd,FMT_PNG,0,fn, filesize, 0);
	}
#endif
#if PV_JPEG_SUPPORT
	else if(fbuf[0]==255 && fbuf[1]==216) {
		load_image_main(hwnd,FMT_JPEG,0,fn, filesize, 0);
	}
#endif
#if PV_GIF_SUPPORT
	else if(fbuf[0]=='G' && fbuf[1]=='I' && fbuf[2]=='F') {
		load_image_main(hwnd,FMT_GIF,0,fn, filesize, 0);
	}
#endif
#if PV_TIFFREAD_SUPPORT
	else if( (fbuf[0]=='I' && fbuf[1]=='I') ||
		     (fbuf[0]=='M' && fbuf[1]=='M') )
	{
		load_image_main(hwnd,FMT_TIFF,0,fn, filesize, 0);
	}
#endif
	else {
		ret=0;
		pv_mesg(hwnd,MSG_E,_T("File type not supported"));
		return 0;
	}

	return 1;
}


// Called after a background color change or similar
// change. If a PNG image is being viewed, it may
// be necessary to reload the image from disk.
static void repaint_things(HWND hwnd)
{
	if(iinfo->img_fmt==FMT_PNG && lstrlen(iinfo->img_filename)) {
		load_image_main(hwnd,FMT_PNG,0,iinfo->img_filename,iinfo->file_size, 1);
	}
	else if(iinfo->img_fmt==FMT_BMP && lstrlen(iinfo->img_filename) && 
		iinfo->bitsperpixel==32)
	{
		load_image_main(hwnd,FMT_BMP,iinfo->img_source,iinfo->img_filename,iinfo->file_size, 1);
	}

	InvalidateRect(hwnd,NULL,FALSE);
}

static int open_image_dialog(HWND hwnd)
{
	TCHAR fn[MAX_PATH];
	OPENFILENAME ofn;

	StringCbCopy(fn,sizeof(fn),_T(""));

	ZeroMemory(&ofn,sizeof(OPENFILENAME));

	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=hwnd;

	// fixme: This should only list supported types
	ofn.lpstrFilter=
		_T("All supported image types\0*.png;*.bmp;*.jpg\0")
		_T("BMP images\0*.bmp\0")
		_T("JPEG images\0*.jpg\0")
		_T("PNG images\0*.png\0")
		_T("All files\0*.*\0\0");

	ofn.nFilterIndex=1;
	ofn.lpstrTitle=_T("Open image file");
	
	ofn.lpstrInitialDir=NULL;
	ofn.lpstrFile=fn;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

	if(!GetOpenFileName(&ofn)) {
		return 0;  /* canceled by user */
	}
	return open_image_by_name(hwnd,fn);
}

static void dropped_files(HWND hwnd, HDROP hDrop)
{
	UINT ret;
	DWORD attr;
	TCHAR fn[MAX_PATH];

	ret=DragQueryFile(hDrop,0xFFFFFFFF,NULL,0);
	if(ret==1) {  	/* number of files dropped */
		ret=DragQueryFile(hDrop,0,fn,MAX_PATH);
		if(ret) {
			attr=GetFileAttributes(fn);
			if((attr==0xffffffff) || !(attr&FILE_ATTRIBUTE_DIRECTORY)) {
				open_image_by_name(hwnd,fn);
			}
		}
	}
	DragFinish(hDrop);
}

// Caller sets initial filename, filename is returned in the
// same buffer.
// returns 0 if user canceled.
int pv_save_as_dialog(HWND hwnd,const TCHAR *title,TCHAR *filename,
					const TCHAR *filter,const TCHAR *defext)
{
	OPENFILENAME ofn;
	BOOL b;

	ZeroMemory(&ofn,sizeof(OPENFILENAME));
	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=hwnd;
	ofn.lpstrFilter=filter; //"PNG image (*.png)\0*.png\0\0";
	ofn.nFilterIndex=1;
	ofn.lpstrTitle=title; //"Save As PNG";
	ofn.lpstrInitialDir=NULL;
	ofn.lpstrFile=filename;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_PATHMUSTEXIST|OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT;
	ofn.lpstrDefExt=defext; //"png";

	b=GetSaveFileName(&ofn);
	return b?1:0;
}


int WINAPI _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
   LPTSTR lpCmdLine,int nCmdShow)
{
	MSG msg;
	WNDCLASS  wc;
	HACCEL hAccTable;
	HWND hwnd;
	int i;

	gl.hInst=hInstance;
	hwnd=NULL;

	einfo = (struct einfo_struct *)calloc(sizeof(struct einfo_struct),1);
	if(!einfo) return 0;

	einfo->heap = GetProcessHeap();
	einfo->software_name = g_software_name;
	einfo->software_name_full = g_software_name_full;
	einfo->software_name_full_a = g_software_name_full_a;
	einfo->monitor_gamma = 2.2;

	iinfo = (struct iinfo_struct *)calloc(sizeof(struct iinfo_struct),1);
	if(!iinfo) return 0;

	iinfo->img_filename = gl.image_filename; // a permanent pointer
	StringCchCopy(iinfo->img_filename,MAX_PATH,_T(""));

	read_settings();  // inits globals

	// lame-o command-line argument processing.
	// We handle one file only, with or without quotes.
	if(lpCmdLine[0]) {
		if(lpCmdLine[0]=='"') {
			StringCchCopy(iinfo->img_filename,MAX_PATH,&lpCmdLine[1]);
			i=lstrlen(iinfo->img_filename);
			if(iinfo->img_filename[i-1]=='"') iinfo->img_filename[i-1]='\0';
		}
		else {
			StringCchCopy(iinfo->img_filename,MAX_PATH,lpCmdLine);
		}
	}

	wc.style = 0;
	wc.lpfnWndProc = (WNDPROC)WndProcGammaDisplay;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = gl.hInst;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName =  NULL;
	wc.lpszClassName = _T("PVGAMMACLASS");
	if(!RegisterClass(&wc)) {
		return 0;
	}

	wc.style = CS_HREDRAW|CS_VREDRAW;
	wc.lpfnWndProc = (WNDPROC)WndProcMain;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = gl.hInst;
	wc.hIcon = LoadIcon(gl.hInst,_T("ICONMAIN"));
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName =  _T("MENUMAIN");
	wc.lpszClassName = _T("PURPLEVIEWCLASS");
	if(!RegisterClass(&wc)) {
		return 0;
	}

	hwnd = CreateWindow(
		_T("PURPLEVIEWCLASS"),einfo->software_name,
		WS_VISIBLE|WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN,
		CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,
		NULL,NULL,gl.hInst,NULL);
	if (!hwnd) return 0;

	DragAcceptFiles(hwnd,TRUE);
	hAccTable=LoadAccelerators(gl.hInst,_T("ACCELTABLE"));

	if(lstrlen(iinfo->img_filename)) {
		open_image_by_name(hwnd,iinfo->img_filename);
	}

	while(GetMessage(&msg,NULL,0,0)) {
		if (!TranslateAccelerator(hwnd,hAccTable,&msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	free((void*)einfo);
	free((void*)iinfo);
	return (msg.wParam);
}

static LRESULT CALLBACK WndProcMain(HWND hwnd, UINT msg,
     WPARAM wParam, LPARAM lParam)
{
	WORD id;
	int x;

	id=LOWORD(wParam);

	switch(msg) {

	case WM_CREATE:

		ShowScrollBar(hwnd,SB_BOTH,gl.shrinkmode?FALSE:TRUE);
		EnableScrollBar(hwnd,SB_BOTH,ESB_DISABLE_BOTH);
		return 0;

	case WM_SIZE:
		set_scrollbars(hwnd);
		return 0;

	case WM_HSCROLL:
		scrollbarmsg_x(hwnd,LOWORD(wParam),HIWORD(wParam));
		return 0;
	case WM_VSCROLL:
		scrollbarmsg_y(hwnd,LOWORD(wParam),HIWORD(wParam));
		return 0;

	case WM_LBUTTONDOWN:
		if(!gl.shrinkmode && gl.lpdib) {
			gl.mousepos=MAKEPOINTS(lParam);
			gl.dragging=1;
			gl.dragpos_x = gl.xscrollpos;
			gl.dragpos_y = gl.yscrollpos;
			SetCapture(hwnd);
			SetCursor(gl.cur_drag2);
			return 0;
		}
		break;

	case WM_LBUTTONUP:
		if(gl.dragging) {
			gl.dragging=0;
			ReleaseCapture();
			return 0;
		}
		break;
	case WM_MOUSEMOVE:
		if(gl.dragging) {
			POINTS pp;
			pp=MAKEPOINTS(lParam);
			mousedrag(hwnd,gl.dragpos_x+gl.mousepos.x-pp.x,
				gl.dragpos_y+gl.mousepos.y-pp.y);
			return 0;
		}
		break;

	case WM_KEYDOWN:
		switch((int)wParam) {
		case VK_LEFT:
			scrollbarmsg_x(hwnd,SB_LINELEFT,0); return 0;
		case VK_RIGHT:
			scrollbarmsg_x(hwnd,SB_LINERIGHT,0); return 0;
		case VK_UP:
			scrollbarmsg_y(hwnd,SB_LINEUP,0); return 0;
		case VK_DOWN:
			scrollbarmsg_y(hwnd,SB_LINEDOWN,0); return 0;
		case VK_END:
			scrollbarmsg_x(hwnd,SB_RIGHT,0); return 0;
		case VK_HOME:
			scrollbarmsg_x(hwnd,SB_LEFT,0); return 0;
		case VK_PRIOR:
			scrollbarmsg_y(hwnd,SB_TOP,0); return 0;
		case VK_NEXT:
			scrollbarmsg_y(hwnd,SB_BOTTOM,0); return 0;
		}
		break;

	case WM_ERASEBKGND:
		// Never erase the background. We always paint the full window
		// in WM_PAINT.
		return 1;

	case WM_PAINT:
		{
			PAINTSTRUCT paintStruct;
			HDC hdc;
			RECT rect, fr;
			//int oldstretchbltmode;
			int xpos,ypos;
			HBRUSH oldbr,br;
			int rv;
			int dst_width,dst_height;
			int wnd_width,wnd_height;
			int delbrushflag;
			double rw,rh;


			hdc = BeginPaint( hwnd, &paintStruct );

			GetClientRect(hwnd,&rect);

			wnd_width=rect.right;
			wnd_height=rect.bottom;

			delbrushflag=0;
			if(einfo->use_image_bg && iinfo->has_bgcolor) {
				br=CreateSolidBrush(iinfo->bgcolor);
				delbrushflag=1;
			}
			else {
				if(einfo->use_custom_bg || gl.bordercolorflag) {
					br=CreateSolidBrush(einfo->custom_bg);
					delbrushflag=1;
				}
				else {
					br=GetSysColorBrush(COLOR_WINDOW);
				}
			}
			oldbr=SelectObject(hdc,br);
			
			if(gl.lpdib) {

				SetViewportOrgEx(hdc,-gl.xscrollpos, -gl.yscrollpos, NULL);

				if( (gl.stretchmode && (iinfo->adj_src_width<wnd_width && iinfo->adj_src_height<wnd_height)) ||
					(gl.shrinkmode  && (iinfo->adj_src_width>wnd_width || iinfo->adj_src_height>wnd_height)))
				{
					rw=(double)wnd_width/iinfo->adj_src_width_flt;
					rh=(double)wnd_height/iinfo->adj_src_height_flt;
					if(rw>rh) {
						dst_width = (int)(iinfo->adj_src_width_flt*rh);
						dst_height=wnd_height;
					}
					else {
						dst_height = (int)(iinfo->adj_src_height_flt*rw);
						dst_width=wnd_width;
					}
				}
				else {
					dst_width=  iinfo->adj_src_width;
					dst_height= iinfo->adj_src_height;
				}

				if(GetStretchBltMode(hdc)!=gl.stretchbltmode) {
					SetStretchBltMode(hdc,gl.stretchbltmode);
					if(gl.stretchbltmode==HALFTONE) {
						// This may not be the right way to do this..
						SetBrushOrgEx(hdc,0,0,NULL);
					}
				}

				xpos=0; ypos=0;
				if(gl.halign==1) {
					if(wnd_width>dst_width)
						xpos=(wnd_width-dst_width)/2;
				}
				else if(gl.halign==2) {
					if(wnd_width>dst_width)
						xpos=wnd_width-dst_width;
				}

				if(gl.valign==1) {
					if(wnd_height>dst_height)
						ypos=(wnd_height-dst_height)/2;
				}
				else if(gl.valign==2) {
					if(wnd_height>dst_height)
						ypos=wnd_height-dst_height;
				}

				rv=StretchDIBits(hdc,
					xpos,ypos, // dest pos
					  dst_width,dst_height,
					0,0,                         // source
					  iinfo->src_width, iinfo->src_height,    
					gl.lpbits,(LPBITMAPINFO)gl.lpdib,DIB_RGB_COLORS,SRCCOPY);

				//if(oldstretchbltmode) SetStretchBltMode(hdc,oldstretchbltmode);

				// Paint the background color only in the area *around*
				// the image. This prevents flicker when the window is
				// resized.
				if(xpos-gl.xscrollpos>0) {    // left part
					fr.left=gl.xscrollpos;       fr.top=gl.yscrollpos;
					fr.right=gl.xscrollpos+xpos; fr.bottom=gl.yscrollpos+rect.bottom;
					FillRect(hdc,&fr,br);
				}
				if(xpos+dst_width-gl.xscrollpos < rect.right) {  // right part
					fr.left=gl.xscrollpos+xpos+dst_width; fr.top=gl.yscrollpos;
					fr.right=gl.xscrollpos+ rect.right;     fr.bottom=gl.yscrollpos+rect.bottom;
					FillRect(hdc,&fr,br);
				}
				if(ypos-gl.yscrollpos>0) {    // top part
					fr.left=gl.xscrollpos+xpos;  fr.top=gl.yscrollpos;
					fr.right=gl.xscrollpos+xpos+dst_width; fr.bottom=gl.yscrollpos+ypos;
					FillRect(hdc,&fr,br);
				}
				if(ypos+dst_height-gl.yscrollpos < rect.bottom) { // bottom part
					fr.left=gl.xscrollpos+xpos;  fr.top=gl.yscrollpos+ypos+dst_height;
					fr.right=gl.xscrollpos+xpos+dst_width; fr.bottom=gl.yscrollpos+rect.bottom;
					FillRect(hdc,&fr,br);
				}

				if(rv==GDI_ERROR || rv<1) {
					SetViewportOrgEx(hdc,0,0, NULL);
					FillRect(hdc,&rect,br);
					TextOut(hdc,0,0,_T("GDI ERROR"),9);
				}
			}
			else {
				FillRect(hdc,&rect,br);
			}

			if(oldbr) {
				SelectObject(hdc,oldbr);
				if(delbrushflag) DeleteObject(br);
			}

			EndPaint( hwnd, &paintStruct );
			return 0;
		}

	case WM_DESTROY:
		save_settings(hwnd);
		if(gl.lpdib) { HeapFree(einfo->heap,0,gl.lpdib); gl.lpdib=NULL; }
		PostQuitMessage(0);
		return 0;

	case WM_DROPFILES:
		dropped_files(hwnd, (HDROP)wParam);
		return 0;

	case WM_INITMENU:
		EnableMenuItem((HMENU)wParam, ID_IMAGEINFO, MF_BYCOMMAND|
			(gl.lpdib?MF_ENABLED:MF_GRAYED) );
		EnableMenuItem((HMENU)wParam, ID_COPY, MF_BYCOMMAND|
			(gl.lpdib?MF_ENABLED:MF_GRAYED) );
		EnableMenuItem((HMENU)wParam, ID_PASTE, MF_BYCOMMAND|
			(IsClipboardFormatAvailable(CF_DIB))?MF_ENABLED:MF_GRAYED);
		CheckMenuItem((HMENU)wParam, ID_STRETCH , MF_BYCOMMAND|
			(gl.stretchmode?MF_CHECKED:MF_UNCHECKED) );
		CheckMenuItem((HMENU)wParam, ID_SHRINK, MF_BYCOMMAND|
			(gl.shrinkmode?MF_CHECKED:MF_UNCHECKED) );
		CheckMenuItem((HMENU)wParam, ID_BG_BKGD , MF_BYCOMMAND|
			(einfo->use_image_bg?MF_CHECKED:MF_UNCHECKED) );
		CheckMenuItem((HMENU)wParam, ID_BG_CUSTOM , MF_BYCOMMAND|
			(einfo->use_custom_bg?MF_CHECKED:MF_UNCHECKED) );
		CheckMenuItem((HMENU)wParam, ID_GAMMACORRECT , MF_BYCOMMAND|
			(einfo->use_gamma?MF_CHECKED:MF_UNCHECKED) );
		EnableMenuItem((HMENU)wParam, ID_CLOSE, MF_BYCOMMAND|
			(gl.lpdib?MF_ENABLED:MF_GRAYED) );
		EnableMenuItem((HMENU)wParam, ID_SAVEASPNG, MF_BYCOMMAND|
			((gl.lpdib&&PV_PNG_SUPPORT)?MF_ENABLED:MF_GRAYED) );
		EnableMenuItem((HMENU)wParam, ID_SAVEASBMP, MF_BYCOMMAND|
			(gl.lpdib?MF_ENABLED:MF_GRAYED) );
		EnableMenuItem((HMENU)wParam, ID_SAVEASTIFF, MF_BYCOMMAND|
			(gl.lpdib?MF_ENABLED:MF_GRAYED) );
		return 0;

	case WM_COMMAND:
		switch(id) {
		case ID_EXIT:
			DestroyWindow(hwnd);
			return 0;
		case ID_OPEN:
			open_image_dialog(hwnd);
			return 0;
		case ID_SAVEASPNG:
			save_image_main(hwnd,FMT_PNG);
			return 0;
		case ID_SAVEASBMP:
			save_image_main(hwnd,FMT_BMP);
			return 0;
		case ID_SAVEASTIFF:
			save_image_main(hwnd,FMT_TIFF);
			return 0;
		case ID_CLOSE:
			if(gl.lpdib) { HeapFree(einfo->heap,0,gl.lpdib); gl.lpdib=NULL; }
			StringCchCopy(iinfo->img_filename,MAX_PATH,_T(""));
			SetTitle(hwnd);
			//SetWindowText(hwnd,einfo->software_name);
			iinfo->has_bgcolor=0;
			iinfo->img_fmt=0;
			InvalidateRect(hwnd,NULL,FALSE);
			set_scrollbars(hwnd);
			return 0;
		case ID_COPY:
			if(gl.lpdib) {
				CopyToClipboard(hwnd,(void*)gl.lpdib,gl.lpdib_size,CF_DIB);
			}
			break;
		case ID_PASTE:
			load_image_main(hwnd,FMT_BMP, 1, _T("[clipboard]"), 0, 0);
			break;
		case ID_STRETCH:
			gl.stretchmode= !gl.stretchmode;
			InvalidateRect(hwnd,NULL,FALSE);
			return 0;
		case ID_SHRINK:
			gl.shrinkmode= !gl.shrinkmode;
			if(gl.shrinkmode) {
				gl.xscrollpos=0;
				gl.yscrollpos=0;
			}
			else {
				gl.xscrollpos=gl.xdesiredscrpos;
				gl.yscrollpos=gl.ydesiredscrpos;
			}
			ShowScrollBar(hwnd,SB_BOTH,gl.shrinkmode?FALSE:TRUE);
			set_scrollbars(hwnd); // this maybe isn't needed
			return 0;

		case ID_BG_BKGD:
			einfo->use_image_bg = !einfo->use_image_bg;
			repaint_things(hwnd);
			return 0;
		case ID_BG_CUSTOM:
			einfo->use_custom_bg = !einfo->use_custom_bg;
			repaint_things(hwnd);
			return 0;
		case ID_GAMMACORRECT:
			einfo->use_gamma = !einfo->use_gamma;
			SetTitle(hwnd);
			repaint_things(hwnd);
			return 0;
		case ID_SETBG:
			{
				if(choose_color_dialog(hwnd,&einfo->custom_bg)) {
					repaint_things(hwnd);
				}
			}
			return 0;
		case ID_ADVOPTIONS:
			x=einfo->bmp_alpha32;
			if(DialogBox(gl.hInst,_T("ADVOPTIONSDLG"),hwnd,DlgProcAdvOptions)>0) {
				calc_image_size();
				set_scrollbars(hwnd);

				if(x!=einfo->bmp_alpha32 && iinfo->bitsperpixel==32 && einfo->use_custom_bg) {
					repaint_things(hwnd);
				}
				else {
					InvalidateRect(hwnd,NULL,TRUE);
				}
			}
			return 0;
		case ID_SETGAMMA:
			if(DialogBox(gl.hInst,_T("GAMMADLG"),hwnd,DlgProcSetGamma)>0) {
				repaint_things(hwnd);
			}
			return 0;
		case ID_IMAGEINFO:
			if(gl.lpdib) {
				DialogBox(gl.hInst,_T("INFODLG"),hwnd,DlgProcInfo);
			}
			return 0;
		case ID_ABOUT:
			DialogBox(gl.hInst,_T("ABOUTDLG"),hwnd,DlgProcAbout);
			return 0;
		}
		break;
	}

	return DefWindowProc(hwnd, msg, wParam, lParam);
}


BOOL CALLBACK DlgProcAbout(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	TCHAR info[5000];
	TCHAR buf[1000];

	buf[0]='\0';
	id=LOWORD(wParam);
	switch (msg) {
	case WM_INITDIALOG:

		StringCbCopy(info,sizeof(info),
			_T("Purpleview\r\nA simple image viewer\r\n")
			_T("Version ") PVVERSION _T("  ") _T(__DATE__) _T("\r\n\r\n")
			_T("Copyright ") PVSYMBOL_COPYRIGHT _T(" 2002-") PVCOPYRIGHT_YEAR _T(" Jason Summers\r\n")
			_T("http://entropymine.com/jason/purpleview/\r\n\r\n")
			_T("This program is distributed under the terms of ")
			_T("the GNU General Public License. Please read the ")
			_T("file purpleview.txt for more information.\r\n\r\n"));

#if PV_PNG_SUPPORT
		buf[0]='\0';
		pv_png_get_copyright(buf,1000);
		if(buf[0]) {
			StringCbCat(info,sizeof(info),buf);
		}
#endif

#if PV_JPEG_SUPPORT
		buf[0]='\0';
		pv_jpeg_get_copyright(buf,1000);
		if(buf[0]) {
			StringCbCat(info,sizeof(info),buf);
		}
#endif

#if PV_GIF_SUPPORT
		buf[0]='\0';
		pv_gif_get_copyright(buf,1000);
		if(buf[0]) {
			strcat(info,buf);
		}
#endif

		SetDlgItemText(hwnd,IDC_ABOUTINFO,info);
		return 1;
	case WM_COMMAND:
		switch(id) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}


static BOOL CALLBACK DlgProcInfo(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	TCHAR info[4000];
	TCHAR buf[500];
	TCHAR buf2[100];
	DWORD tabs[1];
	int flag;

	id=LOWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:
		if(!gl.lpdib) return 1;
		tabs[0]= 70;
		SendDlgItemMessage(hwnd,IDC_INFOTEXT,EM_SETTABSTOPS,(WPARAM)1,(LPARAM)tabs);
		StringCbCopy(info,sizeof(info),_T(""));
		StringCbPrintf(buf,sizeof(buf),_T("File:\t%s\r\n"),iinfo->img_filename);
		StringCbCat(info,sizeof(info),buf);

		if(iinfo->file_size>0) {
			StringCbPrintf(buf,sizeof(buf),_T("File size:\t%d bytes\r\n"),iinfo->file_size);
		}
		else {
			StringCbCopy(buf,sizeof(buf),_T("File size:\tNot available\r\n"));
		}
		StringCbCat(info,sizeof(info),buf);

		StringCbPrintf(buf,sizeof(buf),_T("File format:\t%s\r\n"),fileformat_name[iinfo->img_fmt]);
		StringCbCat(info,sizeof(info),buf);
		StringCbPrintf(buf,sizeof(buf),_T("Color type:\t%s\r\n"),colortype_name[iinfo->colortype]);
		StringCbCat(info,sizeof(info),buf);
		StringCbPrintf(buf,sizeof(buf),_T("Bits/pixel:\t%d"),iinfo->bitsperpixel);
		StringCbCat(info,sizeof(info),buf);

		flag=0;
		StringCbPrintf(buf,sizeof(buf),_T(""));
		if(iinfo->bits_red) {
			StringCbPrintf(buf2,sizeof(buf2),_T("%sR:%d"),flag?_T(","):_T(" ("),iinfo->bits_red);
			StringCbCat(buf,sizeof(buf),buf2);
			flag=1;
		}
		if(iinfo->bits_green) {
			StringCbPrintf(buf2,sizeof(buf2),_T("%sG:%d"),flag?_T(","):_T(" ("),iinfo->bits_green);
			StringCbCat(buf,sizeof(buf),buf2);
			flag=1;
		}
		if(iinfo->bits_blue) {
			StringCbPrintf(buf2,sizeof(buf2),_T("%sB:%d"),flag?_T(","):_T(" ("),iinfo->bits_blue);
			StringCbCat(buf,sizeof(buf),buf2);
			flag=1;
		}
		if(iinfo->bits_gray) {
			StringCbPrintf(buf2,sizeof(buf2),_T("%sGray:%d"),flag?_T(","):_T(" ("),iinfo->bits_gray);
			StringCbCat(buf,sizeof(buf),buf2);
			flag=1;
		}
		if(iinfo->bits_alpha) {
			StringCbPrintf(buf2,sizeof(buf2),_T("%sAlpha:%d"),flag?_T(","):_T(" ("),iinfo->bits_alpha);
			StringCbCat(buf,sizeof(buf),buf2);
			flag=1;
		}
		StringCbCat(buf,sizeof(buf),flag?_T(")\r\n"):_T("\r\n"));
		StringCbCat(info,sizeof(info),buf);

		if(iinfo->colortype==COLORTYPE_PALETTE) {
			StringCbPrintf(buf,sizeof(buf),_T("Colors in palette:\t%d\r\n"),iinfo->palettecolors);
			StringCbCat(info,sizeof(info),buf);
		}

		StringCbPrintf(buf,sizeof(buf),_T("Compression:\t%s\r\n"),compr_name[iinfo->compression]);
		StringCbCat(info,sizeof(info),buf);
		StringCbPrintf(buf,sizeof(buf),_T("Interlaced:\t%s\r\n"),iinfo->interlaced?_T("Yes"):_T("No"));
		StringCbCat(info,sizeof(info),buf);
		StringCbPrintf(buf,sizeof(buf),_T("Dimensions:\t%d x %d pixels\r\n"),get_dib_width(), get_dib_height());
		StringCbCat(info,sizeof(info),buf);

		if(iinfo->res_flag==1) { // 1==image units are in meters
			switch(gl.densityunits) {  // this the units the user wants to use for display
			case 1:
				StringCbPrintf(buf,sizeof(buf),_T("Density:\t%.2f x %.2f pixels/cm\r\n"),(double)(iinfo->res_x/100.0),(double)(iinfo->res_y/100.0));
				break;
			case 2:
				StringCbPrintf(buf,sizeof(buf),_T("Density:\t%.1f x %.1f pixels/inch\r\n"),(double)((iinfo->res_x*2.54)/100.0),(double)((iinfo->res_y*2.54)/100.0));
				break;
			default:
				StringCbPrintf(buf,sizeof(buf),_T("Density:\t%d x %d pixels/meter\r\n"),iinfo->res_x,iinfo->res_y);
			}
		}
		else if(iinfo->res_flag==2) { // image units are unspecified
			StringCbPrintf(buf,sizeof(buf),_T("Density:\t%d x %d pixels/unit\r\n"),iinfo->res_x,iinfo->res_y);
		}
		else {
			StringCbPrintf(buf,sizeof(buf),_T("Density:\tUnknown\r\n"));
		}
		StringCbCat(info,sizeof(info),buf);

		StringCbPrintf(buf,sizeof(buf),_T("\r\nCurrent size:\t%d bytes\r\n"),gl.lpdib_size);
		StringCbCat(info,sizeof(info),buf);
		StringCbPrintf(buf,sizeof(buf),_T("Current bits/pixel:\t%d"),get_dib_bpp());
		StringCbCat(info,sizeof(info),buf);

		SetDlgItemText(hwnd,IDC_INFOTEXT,info);
		return 1;
	case WM_COMMAND:
		switch(id) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}

static BOOL CALLBACK DlgProcAdvOptions(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	int c;

	id=LOWORD(wParam);
	switch (msg) {
	case WM_INITDIALOG:
		CheckDlgButton(hwnd,IDC_ALIGN_TL,(gl.halign==0 && gl.valign==0)?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_ALIGN_T ,(gl.halign==1 && gl.valign==0)?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_ALIGN_TR,(gl.halign==2 && gl.valign==0)?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_ALIGN_L ,(gl.halign==0 && gl.valign==1)?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_ALIGN_C ,(gl.halign==1 && gl.valign==1)?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_ALIGN_R ,(gl.halign==2 && gl.valign==1)?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_ALIGN_BL,(gl.halign==0 && gl.valign==2)?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_ALIGN_B ,(gl.halign==1 && gl.valign==2)?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_ALIGN_BR,(gl.halign==2 && gl.valign==2)?BST_CHECKED:BST_UNCHECKED);

		CheckDlgButton(hwnd,gl.bordercolorflag?IDC_BORDER_CUSTOM:IDC_BORDER_SYSTEM,BST_CHECKED);
		CheckDlgButton(hwnd,IDC_FIXRATIO,gl.fixratio?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_SAVESCROLLPOS,gl.savescrollpos?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hwnd,IDC_BMP_ALPHA32,einfo->bmp_alpha32?BST_CHECKED:BST_UNCHECKED);

		switch(gl.densityunits) {
		case 1:  c=IDC_DENSITY_CM; break;
		case 2:  c=IDC_DENSITY_IN; break;
		default: c=IDC_DENSITY_M;
		}
		CheckDlgButton(hwnd,c,BST_CHECKED);

		switch(gl.stretchbltmode) {
		case BLACKONWHITE: c=IDC_STR_BLACK;    break;
		case WHITEONBLACK: c=IDC_STR_WHITE;    break;
		case HALFTONE:     c=IDC_STR_HALFTONE; break;
		default:           c=IDC_STR_COLOR;
		}
		CheckDlgButton(hwnd,c,BST_CHECKED);

		return 1;

	case WM_COMMAND:
		switch(id) {
		case IDOK:
			if(IsDlgButtonChecked(hwnd,IDC_ALIGN_TL)==BST_CHECKED) { gl.halign=0; gl.valign=0; }
			if(IsDlgButtonChecked(hwnd,IDC_ALIGN_T )==BST_CHECKED) { gl.halign=1; gl.valign=0; }
			if(IsDlgButtonChecked(hwnd,IDC_ALIGN_TR)==BST_CHECKED) { gl.halign=2; gl.valign=0; }
			if(IsDlgButtonChecked(hwnd,IDC_ALIGN_L )==BST_CHECKED) { gl.halign=0; gl.valign=1; }
			if(IsDlgButtonChecked(hwnd,IDC_ALIGN_C )==BST_CHECKED) { gl.halign=1; gl.valign=1; }
			if(IsDlgButtonChecked(hwnd,IDC_ALIGN_R )==BST_CHECKED) { gl.halign=2; gl.valign=1; }
			if(IsDlgButtonChecked(hwnd,IDC_ALIGN_BL)==BST_CHECKED) { gl.halign=0; gl.valign=2; }
			if(IsDlgButtonChecked(hwnd,IDC_ALIGN_B )==BST_CHECKED) { gl.halign=1; gl.valign=2; }
			if(IsDlgButtonChecked(hwnd,IDC_ALIGN_BR)==BST_CHECKED) { gl.halign=2; gl.valign=2; }

			if(IsDlgButtonChecked(hwnd,IDC_BORDER_SYSTEM)==BST_CHECKED) { gl.bordercolorflag=0; }
			if(IsDlgButtonChecked(hwnd,IDC_BORDER_CUSTOM)==BST_CHECKED) { gl.bordercolorflag=1; }

			gl.fixratio = (IsDlgButtonChecked(hwnd,IDC_FIXRATIO)==BST_CHECKED);
			gl.savescrollpos = (IsDlgButtonChecked(hwnd,IDC_SAVESCROLLPOS)==BST_CHECKED);
			einfo->bmp_alpha32 = (IsDlgButtonChecked(hwnd,IDC_BMP_ALPHA32)==BST_CHECKED);

			if(IsDlgButtonChecked(hwnd,IDC_DENSITY_M )==BST_CHECKED) { gl.densityunits=0; }
			if(IsDlgButtonChecked(hwnd,IDC_DENSITY_CM)==BST_CHECKED) { gl.densityunits=1; }
			if(IsDlgButtonChecked(hwnd,IDC_DENSITY_IN)==BST_CHECKED) { gl.densityunits=2; }

			if(IsDlgButtonChecked(hwnd,IDC_STR_COLOR   )==BST_CHECKED) { gl.stretchbltmode=COLORONCOLOR; }
			if(IsDlgButtonChecked(hwnd,IDC_STR_BLACK   )==BST_CHECKED) { gl.stretchbltmode=BLACKONWHITE; }
			if(IsDlgButtonChecked(hwnd,IDC_STR_WHITE   )==BST_CHECKED) { gl.stretchbltmode=WHITEONBLACK; }
			if(IsDlgButtonChecked(hwnd,IDC_STR_HALFTONE)==BST_CHECKED) { gl.stretchbltmode=HALFTONE; }

			EndDialog(hwnd, 1);
			return 1;

		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}


// mode=0: PAINT everything
// mode=1: draw only the parts that change frequently
static void PaintGammaDisplay(int mode, HWND hwnd, HDC hdc,double gamma)
{
	HBRUSH blackbrush;
	HPEN pens[4];
	HBRUSH gbrushes[4];
	HPEN oldpen;
	RECT clientrect;

	int width, height;
	int colorheight;
	int midxoffs, midyoffs;
	int midwidth, midheight;
	int v;
	RECT r;

	int i,k;

	GetClientRect(hwnd,&clientrect);

	if(mode==1) {
		hdc=GetDC(hwnd);
	}

	if(mode==0) {
		oldpen=SelectObject(hdc,GetStockObject(NULL_PEN));
	}

	width=clientrect.right;
	colorheight=clientrect.bottom/4;
	height=colorheight*4;
	midxoffs=30; midwidth=width-2*midxoffs;
	midyoffs=18; midheight=colorheight-2*midyoffs;

	if(mode==0) {
		blackbrush=CreateSolidBrush(RGB(0,0,0));
		pens[0]=CreatePen(PS_SOLID,0,RGB(255,0,0));
		pens[1]=CreatePen(PS_SOLID,0,RGB(0,255,0));
		pens[2]=CreatePen(PS_SOLID,0,RGB(0,0,255));
		pens[3]=CreatePen(PS_SOLID,0,RGB(255,255,255));
	}

	v=(unsigned char)(pow(0.5,(1.0/gamma))*255.0+0.5);

	gbrushes[0]=CreateSolidBrush(RGB(v,0,0));
	gbrushes[1]=CreateSolidBrush(RGB(0,v,0));
	gbrushes[2]=CreateSolidBrush(RGB(0,0,v));
	gbrushes[3]=CreateSolidBrush(RGB(v,v,v));

	if(mode==0) {
		FillRect(hdc,&clientrect,blackbrush);
	}

	for(k=0;k<4;k++) {
		if(mode==0) {
			SelectObject(hdc,(HGDIOBJ)pens[k]);

			for(i=0;i<colorheight;i+=2) {
				MoveToEx(hdc,0,i+colorheight*k,NULL);
				LineTo(hdc,width,i+colorheight*k);
			}
		}

		r.left=midxoffs;
		r.right=r.left+midwidth;
		r.top=midyoffs+k*colorheight;
		r.bottom=r.top+midheight;

		FillRect(hdc,&r,gbrushes[k]);
	}

	if(mode==0) {
		SelectObject(hdc,oldpen);
		DeleteObject((HGDIOBJ)blackbrush);
	}

	for(k=0;k<4;k++) {
		if(mode==0) {
			DeleteObject((HGDIOBJ)pens[k]);
		}
		DeleteObject((HGDIOBJ)gbrushes[k]);
	}

	if(mode==1) {
		ReleaseDC(hwnd,hdc);
	}
}

static void UpdateGamma(HWND hwndDlg, HWND hwndGamma, int needsetpos, int seteditctrl)
{
	TCHAR s[80];

	if(seteditctrl) {
		StringCbPrintf(s,sizeof(s),_T("%.2f"),gl.tmp_displayed_gamma);
		SetDlgItemText(hwndDlg,IDC_GAMMA,s);
	}

	if(needsetpos) {
		SendDlgItemMessage(hwndDlg,IDC_GAMMASLIDER,TBM_SETPOS,
			(WPARAM)TRUE,(LPARAM)(LONG)(gl.tmp_displayed_gamma*50.0+0.5));
	}

	PaintGammaDisplay(1,hwndGamma,NULL,gl.tmp_displayed_gamma);
}

static double GetGammaFromEditCtrl(HWND hwndDlg)
{
	double g;
	TCHAR s[80];

	GetDlgItemText(hwndDlg,IDC_GAMMA,s,80);
	g = _tstof(s);
	if(g<0.02) g=0.02;
	if(g>20.0) g=20.0;
	return g;
}


static void GetPosInParent(HWND hwnd,RECT *rc)
{
	POINT p;

	p.x= p.y= 0;
	ClientToScreen(GetParent(hwnd),&p);
	GetWindowRect(hwnd,rc);
	rc->left -= p.x;
	rc->right -= p.x;
	rc->top -= p.y;
	rc->bottom -= p.y;
}


static BOOL CALLBACK DlgProcSetGamma(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id,code;
	int i;
	RECT r1,r2,r3;
	static HWND hwndgammadisplay;
	int height;

	id=LOWORD(wParam);
	code=HIWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:

		gl.tmp_displayed_gamma=einfo->monitor_gamma;

		SetDlgItemText(hwnd,IDC_INSTR,
			_T("Adjust the slider until the central solid colors are the ")
			_T("same brightness as the surrounding areas.\r\n\r\n")
			_T("It's usually best to leave this at the default 2.20."));


		GetPosInParent(GetDlgItem(hwnd,IDC_GAMMA),&r1);
		GetPosInParent(GetDlgItem(hwnd,IDC_GAMMASLIDER),&r2);
		GetPosInParent(GetDlgItem(hwnd,IDOK),&r3);

		height = r2.top-r3.top-7;
		height = (height/8)*8;

		hwndgammadisplay = CreateWindow(_T("PVGAMMACLASS"),_T(""),
			WS_VISIBLE|WS_CHILD,
			r2.left,r3.top,
			r1.left-r2.left-12,height,
			hwnd,NULL,gl.hInst,NULL);

		SendDlgItemMessage(hwnd,IDC_GAMMASLIDER,TBM_SETRANGE,(WPARAM)FALSE,
			(LPARAM)MAKELONG(10,250));

		SendDlgItemMessage(hwnd,IDC_GAMMASLIDER,TBM_SETPAGESIZE,0,(LPARAM)25);

		for(i=50;i<250;i+=50) {
			SendDlgItemMessage(hwnd,IDC_GAMMASLIDER,TBM_SETTIC,0,(LPARAM)(LONG)i);
		}

		UpdateGamma(hwnd,hwndgammadisplay,1,1);

		return 1;

	case WM_HSCROLL:
		{
			// We assume this message is from the slider control, since it is
			// the only control that should be sending scroll messages.
			int nScrollCode;
			short int nPos;

			nScrollCode = (int) LOWORD(wParam);
			nPos = (short int) HIWORD(wParam);

			switch(nScrollCode) {
			case TB_TOP:
			case TB_BOTTOM:
			case TB_LINEUP:
			case TB_LINEDOWN:
			case TB_PAGEUP:
			case TB_PAGEDOWN:
				i=SendDlgItemMessage(hwnd,IDC_GAMMASLIDER,TBM_GETPOS,0,0);
				gl.tmp_displayed_gamma = ((double)i)/50.0;
				UpdateGamma(hwnd,hwndgammadisplay,0,1);
				break;

			case TB_THUMBTRACK:
				gl.tmp_displayed_gamma = ((double)nPos)/50.0;
				UpdateGamma(hwnd,hwndgammadisplay,0,1);
				break;
			}

		}
		return 1;

	case WM_COMMAND:
		switch(id) {
		case IDC_GAMMA:
			if(code==EN_CHANGE) {
				gl.tmp_displayed_gamma = GetGammaFromEditCtrl(hwnd);
				UpdateGamma(hwnd,hwndgammadisplay,1,0);
				return 1;
			}
			return 0;

		case IDC_DEFAULTGAMMA:
			gl.tmp_displayed_gamma = 2.20;
			UpdateGamma(hwnd,hwndgammadisplay,1,1);
			return 1;

		case IDOK:
			einfo->monitor_gamma = GetGammaFromEditCtrl(hwnd);
			EndDialog(hwnd, 1);
			return 1;

		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}

static LRESULT CALLBACK WndProcGammaDisplay(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;


	id=LOWORD(wParam);

	switch(msg) {

	case WM_CREATE:
		return 0;

	case WM_ERASEBKGND:
		return 1;

	case WM_PAINT:
		{
			PAINTSTRUCT paintStruct;
			HDC hdc;
			hdc = BeginPaint( hwnd, &paintStruct );
			PaintGammaDisplay(0, hwnd, hdc,gl.tmp_displayed_gamma);
			EndPaint( hwnd, &paintStruct );
			return 0;
		}

	case WM_DESTROY:
		return 0;
	} 

	return DefWindowProc(hwnd, msg, wParam, lParam);
}
